#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@created: 22.06.20
@author: eisenmenger
"""
import pytest

from unittest import mock
from unittest.mock import call
from kombunicator.tasks import _rpc_server_publisher


@pytest.fixture(scope='module')
def conn_param():
    return {
        'hostname': '127.0.0.1',
        'port': '6397',
        'userid': 'jondoe',
        'password': 'superstrong_password',
    }


@mock.patch('kombunicator.tasks.MessageProducer')
def test_rpc_server_publisher_task(mocked_msg_prod, conn_param):
    def _test_message_producer_init():
        _rpc_server_publisher(message='Some message',
                              connection_parameter=conn_param,
                              routing_key='rpc_server_key',
                              correlation_id='service')

        mocked_msg_prod.assert_called_with(conn_param)
        assert mocked_msg_prod.called
        assert mocked_msg_prod.called_once

    def _test_message_producer_publish_called():
        _rpc_server_publisher(message='Some other message',
                              connection_parameter=conn_param,
                              routing_key='rpc_server_key',
                              correlation_id='operation')

        expected_call = call().publish(message='Some other message',
                                       routing_key='rpc_server_key',
                                       correlation_id='operation')
        assert mocked_msg_prod.mock_calls[-1] == expected_call

    _test_message_producer_init()
    _test_message_producer_publish_called()



import json
import uuid
from threading import Event
from types import FunctionType

from kombu import Connection, Producer
from strongtyping.strong_typing import match_typing

from kombunicator.MessageConsumer import MessageConsumer


class RPCClient:

    @match_typing
    def __init__(self, connection_parameter: dict, request_key: str, default_callback: FunctionType = None):
        self.connection_parameter = connection_parameter
        self.request_key = request_key
        self.default_callback = default_callback

    @property
    def _unique_string(self):
        return str(uuid.uuid4().hex)

    @match_typing(excep_raise=TypeError)
    def request(self, message: dict, headers=None, **kwargs):
        """Requests data from a remote PRC server.

        Parameters
        ----------
        message : dict
            representation of data to be processed remotely
        headers: dict
            additional, optional headers set to the message
        kwargs : dict
            - Can have a 'callback' key with a callable attached.
              Then, the callable will be used as callback to process the
              answer received from the remote RPC server. Otherwise
              the default_callback defined in the __init__() method will
              be used.
            - all other KV pairs of kwargs will be passed
              to the callback as parameters.

        Raises
        ------
            ValueError
        """
        reply_callback = kwargs.pop('callback', self.default_callback)
        if not isinstance(reply_callback, FunctionType):
            raise ValueError('Callback must be a function type.')

        try:
            json.dumps(headers)
        except TypeError:
            raise TypeError('Headers is not JSON serializable')

        message_correlation_id = self._unique_string
        result_queue_name = self._unique_string
        _ready = Event()

        # create a threaded direct message consumer
        self.consumer_thread = MessageConsumer(
            connection_parameter=self.connection_parameter,
            consumer_type='direct',
            binding_keys=[result_queue_name],
            _thread_ready=_ready
        )

        def _callback(payload, headers, properties):
            if properties['correlation_id'] == message_correlation_id:
                reply_callback(payload, **kwargs)
                self.consumer_thread.stop()
            else:
                # TODO: report bad message routing and/or wrong correlation id
                pass

        self.consumer_thread.register_message_handler(_callback)
        self.consumer_thread.start()
        _ready.wait()

        with Connection(**self.connection_parameter) as connection:
            with Producer(connection) as producer:
                producer.publish(
                    body=message,
                    exchange='',
                    routing_key=self.request_key,
                    reply_to=result_queue_name,
                    correlation_id=message_correlation_id,
                    headers=headers
                )




import signal
from threading import Event

from kombunicator.ConsumerConfigurator import ConsumerConfigurator
from kombunicator.MessageProducer import MessageProducer
from kombunicator.RPCClient import RPCClient
from kombunicator.RPCServer import RPCServer

_consumer_container = list()


def shutdown_consumers():
    [t.stop() for t in _consumer_container]
    [t.join() for t in _consumer_container]


def _system_signal_handler(signum, frame):
    shutdown_consumers()


signal.signal(signal.SIGINT, _system_signal_handler)
signal.signal(signal.SIGTERM, _system_signal_handler)


def register_message_consumer(config_class):
    """
    Setup and starts a message consumer thread.

    :type config: Instance of kombunicator.ConsumerConfigurator
    """
    # import only intarnally
    from kombunicator.MessageConsumer import MessageConsumer

    config = config_class()
    _ready = Event()

    consumer_direct_params = {
        'connection_parameter': config.connection_parameter,
        'consumer_type': config.consumer_type,
        'binding_keys': config.binding_keys,
        '_thread_ready': _ready
    }

    if config.consumer_type == 'direct':
        consumer_thread = MessageConsumer(**consumer_direct_params)

    elif config.consumer_type == 'topic':
        consumer_thread = MessageConsumer(
            exchange_name=config.exchange_name,
            q_unique=config.q_unique,
            accept=config.accept,
            **consumer_direct_params
        )
    else:
        raise ValueError('consumer_type must be direct or topic.')

    def callback(obj, payload, headers, properties):
        '''obj will be occupied by self'''
        config.message_handler(payload, headers, properties)

    consumer_thread.register_message_handler(callback)
    try:
        consumer_thread.start()
    except RuntimeError:
        raise
    else:
        _consumer_container.append(consumer_thread)
        _ready.wait()




import uuid

from kombu import Exchange, Queue, binding

from kombunicator.BaseConsumer import BaseConsumer


class TopicConsumer(BaseConsumer):

    def __init__(self, connection, exchange_name: str, binding_keys: list, accept: list = ['json'], q_unique=None):
        super().__init__()
        self.connection = connection
        self.exchange_name = exchange_name
        self.binding_keys = binding_keys
        self.accept = accept
        self.q_unique = q_unique

        self.exchange = Exchange(self.exchange_name, type='topic')
        self.q_bindings = [binding(self.exchange, routing_key=key) for key in self.binding_keys]

        if self.q_unique is None:
            self.q_unique = uuid.uuid4().hex
        q_name = f"{exchange_name}:{'|'.join([str(key) for key in binding_keys])}:{q_unique}"
        if len(q_name) > 255:
            raise ValueError("Queue name too long. Maximal 255 characters allowed")

        self.queue = Queue(
            name=q_name,
            exchange=self.exchange,
            bindings=self.q_bindings,
            auto_delete=True,
            expires=2 * 30 * 24 * 60 * 60
        )



import logging
from datetime import datetime as dt
from typing import Any

from kombu.message import MessageStateError
from kombu.mixins import ConsumerMixin
from kombu.transport.pyamqp import Message


class BaseConsumer(ConsumerMixin):

    def __init__(self):
        self._add_handler_to_logger()
        super().__init__()

    def get_consumers(self, Consumer, channel):
        consumers = list()
        consumer = Consumer(queues=self.queue,
                            accept=self.accept,
                            callbacks=[self._handle_message])
        consumers.append(consumer)
        return consumers

    def _handle_message(self, body, message: Message):
        cid = message.properties.get('correlation_id', 'unknown')
        try:
            # Acknowledge this message as being processed.
            message.ack()
        except MessageStateError:  # If the message has already been acknowledged/requeued/rejected.
            self.__create_log_msg__(True, cid, message, body)
        else:
            self.process_message(message.payload, message.headers, message.properties)

    def process_message(self, payload, headers, properties):
        """
        Must be overwritten by subclass.
        """
        raise NotImplementedError('Subclass responsibility.')

    def _add_handler_to_logger(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        fh = logging.FileHandler(f'/tmp/{self.__class__.__name__}.log')
        fh.setLevel(logging.INFO)
        self.logger.addHandler(fh)

    def __create_log_msg__(self, error: bool, cid: str, message: Message, body: Any = None):
        msg = '\n\t'.join([f'{k}: {v}' for k, v in message.payload.items()])
        msg += '\n\t'.join([f'{k}: {v}' for k, v in message.headers.items()])
        msg += '\n\t'.join([f'{k}: {v}' for k, v in message.properties.items()])
        if body is not None:
            msg += '\n\t'.join([f'{k}: {v}' for k, v in body.items()])

        log_time = dt.utcnow().replace(microsecond=0)
        if error:
            self.logger.info(f'{log_time}: Message {cid} has already been processed, rejected or requeued.')
        else:
            self.logger.info(f'{log_time}: Message with cid="{cid}" will be processed: \n\t {msg}')




from celery import shared_task
from kombunicator.MessageProducer import MessageProducer

# see: https://stackoverflow.com/questions/48320227/import-shared-tasks-in-celery


@shared_task
def _rpc_server_publisher(message, connection_parameter: dict, routing_key: str, correlation_id: str):
    producer = MessageProducer(connection_parameter)
    producer.publish(
        message=message,
        routing_key=routing_key,
        correlation_id=correlation_id
    )




from threading import Thread
from types import FunctionType, MethodType

from kombu import Connection
from strongtyping.strong_typing import match_typing

from kombunicator.DirectConsumer import DirectConsumer
from kombunicator.TopicConsumer import TopicConsumer


class MessageConsumer(Thread):

    @match_typing
    def __init__(self, connection_parameter: dict = {}, consumer_type: str = 'direct', exchange_name='',
                 binding_keys: list = [], q_unique=None, accept: list = ['json'],
                 _thread_ready=None):
        super().__init__()
        self.daemon = True

        self.connection_parameter = connection_parameter
        self.consumer_type = consumer_type
        self.binding_keys = binding_keys
        self.accept = accept
        self._thread_ready = _thread_ready

        if self.consumer_type == 'topic':
            assert type(exchange_name) is str, f"'exchange_name' must be string"
            assert type(q_unique) is str, f"'q_unique' must be string"

        self.exchange_name = exchange_name
        self.q_unique = q_unique

    def register_message_handler(self, msg_handler):
        assert type(msg_handler) is FunctionType or MethodType, "'msg_handler' must be function or method"
        self.msg_handler = msg_handler

    def _create_consumer(self, connection):

        consumer_base_params = {
            'connection': connection,
            'accept': self.accept
        }

        if self.consumer_type == 'topic':
            self.consumer = TopicConsumer(
                exchange_name=self.exchange_name,
                binding_keys=self.binding_keys,
                q_unique=self.q_unique,
                **consumer_base_params
            )
            self.consumer.process_message = MethodType(self.msg_handler, TopicConsumer)

        elif self.consumer_type == 'direct':
            self.consumer = DirectConsumer(binding_key=self.binding_keys[0],
                                           **consumer_base_params)
            self.consumer.process_message = self.msg_handler

        else:
            raise ValueError("consumer_type must be either 'direct' or 'topic'")

        # set ready event of cnsumer is ready to accept messages
        def _on_consume_ready(connection, channel, consumers, th_ready=self._thread_ready):
            th_ready.set()

        self.consumer.on_consume_ready = _on_consume_ready

    def run(self):
        """
        Starts the RabbitMQ consumer. This call will be blocking.
        """
        with Connection(**self.connection_parameter) as conn:
            self._create_consumer(conn)
            self.consumer.run()

    def stop(self):
        """
        Causes the RabbitMQ consumer to stop consuming
        and return. After that, the thread can be joined.
        """
        self.consumer.should_stop = True




from kombu import Connection, Producer

from kombunicator.Exceptions import AMQPConnectionError


class MessageProducer():
    def __init__(self, connection_parameter: dict):
        """Create a message producer to publish messages
        to a RabbitMQ instance.

        Parameters
        ----------
        connection_parameter : `dict`
            Holds the connection parameters to a RabbitMQ instance.
            - 'hostname': `str` Host where RabbitMQ is accessible
            - 'virtual_host': `str` virtual host of RabbitMQ instance
            - 'port': `int` port to RabbitMQ instance
            - 'userid': `str` username
            - 'password': `str` password

        Raises
        ------
        `kombunicator.AMQPConnectionError`
        """
        self.connection_parameter = connection_parameter

        try:
            conn = Connection(**self.connection_parameter)
            conn.connect()
        except (ConnectionRefusedError, AttributeError, TypeError):
            raise AMQPConnectionError('AMQP connection cannot be established.')
        else:
            # when try was successful
            conn.release()

    def publish(self, message, headers: dict = {}, exchange: str = '',
                routing_key: str = 'default', correlation_id: str = ''):
        """Publish a message to the RabbitMQ instance.

        Parameters
        ----------
        message : `dict` or `str`
            Message to be published
        headers : `dict`
            Mapping of arbitrary headers to pass along
            with the message
        exchange : `str`
            Name of the exchange to publish a message to
        routing_key : `str`
            Queue routing key
        correlation_id : `str`
            correlation ID to match a specific message
        """
        with Connection(**self.connection_parameter) as connection:
            with Producer(connection) as producer:
                producer.publish(
                    body=message,
                    headers=headers,
                    exchange=exchange,
                    routing_key=routing_key,
                    correlation_id=correlation_id,
                    retry_policy={
                        'interval_start': 0,
                        'interval_step': 2,
                        'interval_max': 30,
                        'max_retries': 30
                    }
                )




import json
import pytest
import time
from datetime import datetime
from kombunicator import RPCClient

from test import get_random_string


@pytest.mark.timeout(60)
def test_rpc_with_default_processor(conn_param, tmpdir, rpc_key):
    f_name = f'{tmpdir}/rpc_file.json'

    def answer_processor(data):
        data['answer_received'] = str(datetime.now())
        with open(f_name, 'w') as fh:
            fh.write(json.dumps(data))

    rpc_client = RPCClient(
        connection_parameter=conn_param,
        request_key=rpc_key,
        default_callback=answer_processor
    )

    test_message = {
        'data': get_random_string(),
        'issued': str(datetime.now())
    }
    rpc_client.request(test_message)

    # wait for receptions
    time.sleep(2)

    with open(f_name) as fh:
        received_message = json.loads(fh.read())

    # assert test_arg == argument
    assert received_message['data'] == test_message['data']
    assert 'answer_received' in received_message
    assert 'processed' in received_message


@pytest.mark.timeout(60)
def test_rpc_with_dynamic_processor(conn_param, tmpdir, rpc_key):
    f_name = f'{tmpdir}/rpc_file.json'
    argument = get_random_string()

    rpc_client = RPCClient(
        connection_parameter=conn_param,
        request_key=rpc_key
    )

    test_data = {
        'data': get_random_string(),
        'issued': str(datetime.now())
    }

    # RPC client configuration
    def answer_processor(data, arg=''):
        data['arg'] = arg
        with open(f_name, 'w') as fp:
            fp.write(json.dumps(data))

    rpc_client.request(test_data, callback=answer_processor, arg=argument)

    # wait for receptions
    time.sleep(2)

    with open(f_name) as fp:
        received_message = json.loads(fp.read())

    assert received_message.get('arg', '') == argument
    assert received_message.get('data', '') == test_data['data']
    # from the RPC server
    assert 'processed' in received_message
    # from the answer processor
    assert 'arg' in received_message


@pytest.mark.timeout(60)
def test_rpc_with_headers(conn_param, tmpdir, rpc_key_headers):
    f_name = f'{tmpdir}/rpc_file.json'
    argument = get_random_string()

    rpc_client = RPCClient(
        connection_parameter=conn_param,
        request_key=rpc_key_headers
    )

    test_data = {
        'data': get_random_string(),
        'issued': str(datetime.now())
    }

    # RPC client configuration
    def answer_processor(data, arg=''):
        data['arg'] = arg
        with open(f_name, 'w') as fp:
            fp.write(json.dumps(data))

    headers = {"custom_header_key": "custom_header_value"}
    rpc_client.request(test_data, headers=headers, callback=answer_processor, arg=argument)

    # wait for receptions
    time.sleep(2)

    with open(f_name) as fp:
        received_message = json.loads(fp.read())

    assert received_message.get('arg', '') == argument
    assert received_message.get('data', '') == test_data['data']
    assert received_message.get('headers', '') == headers

    # from the RPC server
    assert 'processed' in received_message
    # from the answer processor
    assert 'arg' in received_message


@pytest.mark.timeout(60)
def test_rpc_headers_must_json_serializable(conn_param, tmpdir, rpc_key_headers):

    f_name = f'{tmpdir}/rpc_file.json'
    argument = get_random_string()

    rpc_client = RPCClient(
        connection_parameter=conn_param,
        request_key=rpc_key_headers
    )

    test_data = {
        'data': get_random_string(),
        'issued': str(datetime.now())
    }

    # RPC client configuration
    def answer_processor(data, arg=''):
        data['arg'] = arg
        with open(f_name, 'w') as fp:
            fp.write(json.dumps(data))

    from datetime import datetime as some_class
    for data in [some_class, b'custom_header_value', lambda x: x ** 2, answer_processor]:
        headers = {"custom_header_key": data}
        with pytest.raises(TypeError):
            rpc_client.request(test_data, headers=headers, callback=answer_processor, arg=argument)


@pytest.mark.timeout(200)
def test_rpc_multi_requests(conn_param, tmpdir, rpc_key):
    number_requests = 10

    rpc_client = RPCClient(
        connection_parameter=conn_param,
        request_key=rpc_key
    )

    toc = dict()
    for i in range(number_requests):
        toc[f'f_name_{i}'] = f'{tmpdir}/rpc_file_{i}.json'
        toc[f'data_{i}'] = get_random_string()
        toc[f'arg_{i}'] = get_random_string()

        def answer_processor(data, arg=''):
            data['arg'] = arg
            with open(toc[f'f_name_{i}'], 'w') as fp:
                fp.write(json.dumps(data))

        test_data = {
            'data': toc[f'data_{i}'],
            'issued': str(datetime.now())
        }

        rpc_client.request(test_data, callback=answer_processor, arg=toc[f'arg_{i}'])
        time.sleep(2)

    # now check the results
    for i in range(number_requests):
        with open(toc[f'f_name_{i}']) as fp:
            received_message = json.loads(fp.read())

        assert received_message.get('data', '') == toc[f'data_{i}']
        assert received_message.get('arg', '') == toc[f'arg_{i}']
        # from the RPC server
        assert 'processed' in received_message




import inspect
import json
from typing import List

from strongtyping.strong_typing import match_typing
from kombunicator.utils import getter_setter

NEEDED_PARAMS = ['_consumer_type', '_connection_parameter', '_exchange_name', '_binding_keys', '_q_unique', '_accept']


class ConsumerConfigurator:

    def __new__(cls, *args, **kwargs):
        if 'NotImplementedError' in inspect.getsource(cls.message_handler):
            raise NotImplementedError('message_handler: Must be overwritten by subclass.')
        return super(ConsumerConfigurator, cls).__new__(cls)

    def __init__(self):
        self.configure()
        self.__check_configure()

    def configure(self):
        """
        This method is called after object creation.
        The following members must be assigned:
        self.consumer_type
        self.connection_parameter
        self.exchange_name
        self.binding_keys
        self.q_unique
        self.accept
        """
        raise NotImplementedError('Must be overwritten by subclass.')

    def __check_configure(self):
        """
        check if really all needed params where overwritten in configure
        """
        if not all(hasattr(self, param) for param in NEEDED_PARAMS):
            raise NotImplementedError('Read docstring of configure method')

    @classmethod
    def message_handler(cls, payload, headers, properties):
        """
        The callback to be executed on a received message.
        This must be overwritten by subclass.
        """
        raise NotImplementedError('Must be overwritten by subclass.')

    @getter_setter
    @match_typing(excep_raise=TypeError)
    def consumer_type(self, _type: str = None):
        """
        Set name of the type from which the consumer should consume.
        """
        if _type is not None:
            self._consumer_type = _type
        return self._consumer_type

    @getter_setter
    @match_typing(excep_raise=TypeError)
    def connection_parameter(self, param: dict = None):
        """
        Set the parameters for establishing a connection
        to a RabbitMQ service.

        :type param: `dict`
        """
        if param is not None:
            self._connection_parameter = param
        return self._connection_parameter

    @getter_setter
    @match_typing(excep_raise=TypeError)
    def exchange_name(self, name: str = None):
        """
        Set name of the subscribed exchange.
        """
        if name is not None:
            if name.startswith("amq."):
                raise ValueError("Exchange names starting with 'amq.' not allowed.")
            self._exchange_name = name
        return self._exchange_name

    @getter_setter
    @match_typing(excep_raise=TypeError)
    def binding_keys(self, keys: List[str] = None):
        """
        Set binding keys for the subscribed topics.
        """
        if keys is not None:
            self._binding_keys = keys
        return self._binding_keys

    @getter_setter
    @match_typing(excep_raise=TypeError)
    def q_unique(self, value: str = None):
        """
        Set queue name addition to make a unique queue naming.
        Default is a UUID4 string.
        """
        if value is not None:
            self._q_unique = value
        return self._q_unique

    @getter_setter
    @match_typing(excep_raise=TypeError)
    def accept(self, value: List[str] = None):
        """
        List of accepted content_types.
        Default is ['json'] only.
        """
        if value is not None:
            self._accept = value
        return self._accept

    def __str__(self):
        result = [f"Connection Parameters:\n    {json.dumps(self.connection_parameter, indent=2)}\n",
                  f"Exchange Name:\n    {self.exchange_name}\n",
                  f"Binding Keys: \n    {self.binding_keys}\n",
                  f"Unique queue: \n    {self.q_unique}\n",
                  f"Accept: \n    {self.accept}\n",
                  f"Message Handler:\n{inspect.getsource(self.message_handler)}"]
        return "".join(result)

    def __repr__(self):
        return self.__str__()




import uuid

from kombu import Queue
from strongtyping.strong_typing import match_typing

from kombunicator.BaseConsumer import BaseConsumer


class DirectConsumer(BaseConsumer):

    @match_typing
    def __init__(self, connection, binding_key: str = '', accept: list = ['json']):
        super().__init__()

        self.connection = connection
        self.accept = accept

        if binding_key == '':
            self.binding_key = uuid.uuid4().hex
        else:
            self.binding_key = binding_key

        self.queue = Queue(
            name=self.binding_key,
            auto_delete=True
        )
